# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from cadccutout.transform import Transform
from cadccutout.pixel_cutout_hdu import PixelCutoutHDU
from cadccutout.no_content_error import NoContentError
from cadccutout.file_helpers.base_file_helper import BaseFileHelper
from cadccutout.utils import is_integer
from astropy.nddata import NoOverlapError
from astropy.wcs import WCS
from astropy.io.fits import PrimaryHDU
from astropy.io import fits


import logging


__all__ = ['FITSHelper']

logger = logging.getLogger(__name__)


class FITSHelper(BaseFileHelper):

    def __init__(self, input_stream, output_writer):
        """
        Create a new BaseFileHelper used for different file types.  Concrete
         instances are expected to extend this
        class to provide some common state.

        :param input_stream:    The Reader to read the file data from.
        :param output_writer:   The Writer to write the cutout to.
        """
        super(FITSHelper, self).__init__(input_stream, output_writer)

    def _post_sanitize_header(self, header, cutout_result):
        """
        Fix the CRPIX offset
        """
        if cutout_result.wcs_crpix is not None:
            cutout_crpix = cutout_result.wcs_crpix

            for idx, val in enumerate(cutout_crpix):
                header.set('CRPIX{}'.format(idx + 1), val)

    def _get_wcs(self, header):
        naxis_value = header.get('NAXIS')

        if naxis_value is not None and int(naxis_value) > 0:
            for nax in range(1, naxis_value + 1):
                ctype = header.get('CTYPE{0} '.format(nax))
                if ctype is not None and ctype.endswith('-SIP'):
                    naxis = 2
                    break
                else:
                    naxis = None
        else:
            naxis = naxis_value

        return WCS(header=header, naxis=naxis, fix=False)

    def _write_cutout(self, header, data, cutout_dimension, wcs):
        cutout_result = self.do_cutout(
            data=data, cutout_dimension=cutout_dimension, wcs=wcs)

        self._post_sanitize_header(header, cutout_result)

        fits.append(filename=self.output_writer, data=cutout_result.data,
                    header=header, overwrite=False,
                    output_verify='exception', checksum='remove')

        self.output_writer.flush()

    def _pixel_cutout(self, header, data, cutout_dimension):
        extension = cutout_dimension.get_extension()
        wcs = self._get_wcs(header)

        try:
            logger.debug(
                'Cutting out from extension {}'.format(extension))
            self._write_cutout(header=header, data=data,
                               cutout_dimension=cutout_dimension, wcs=wcs)
        except NoOverlapError:
            logging.error(
                'No overlap found for extension {}'.format(extension))
            raise NoContentError('No content (arrays do not overlap).')

    def _is_extension_requested(
            self, extension_idx, ext_name_ver, cutout_dimension):
        requested_extension = cutout_dimension.get_extension()

        matches = False

        if ext_name_ver is not None:
            matches = (
                ext_name_ver == requested_extension or requested_extension ==
                ext_name_ver[0])

        if not matches and is_integer(requested_extension) and is_integer(
                extension_idx):
            matches = (int(requested_extension) == int(extension_idx))

        if not matches:
            matches = (requested_extension == extension_idx)

        return matches

    def _check_hdu_list(self, cutout_dimensions, hdu_list):
        has_match = False
        pixel_matches_left = len(cutout_dimensions)
        for curr_extension_idx, hdu in enumerate(hdu_list):
            # If we encounter a PrimaryHDU, write it at the top and continue.
            if isinstance(hdu, PrimaryHDU) and hdu.data is None:
                logger.debug(
                    'Appending Primary from index {}'.format(
                        curr_extension_idx))
                fits.append(
                    filename=self.output_writer, header=hdu.header, data=None,
                    overwrite=False, output_verify='silentfix',
                    checksum='remove')
            elif hdu.is_image:
                header = hdu.header
                ext_name = header.get('EXTNAME')
                ext_ver = header.get('EXTVER', 0)
                curr_ext_name_ver = None

                if ext_name is not None:
                    curr_ext_name_ver = (ext_name, ext_ver)

                try:
                    if isinstance(cutout_dimensions[0], PixelCutoutHDU):
                        for cutout_dimension in cutout_dimensions:
                            if self._is_extension_requested(
                                    curr_extension_idx, curr_ext_name_ver,
                                    cutout_dimension):
                                logger.debug(
                                    '*** Extension {} does match ({} | {})'
                                    .format(
                                        cutout_dimension.get_extension(),
                                        curr_extension_idx, curr_ext_name_ver))
                                pixel_matches_left -= 1
                                self._pixel_cutout(
                                    header, hdu.data, cutout_dimension)
                                has_match = True

                        if pixel_matches_left == 0:
                            return has_match
                    else:
                        logger.debug('Handling WCS transform.')
                        # Handle WCS transform.
                        transform = Transform()
                        transformed_cutout_dimension = \
                            transform.world_to_pixels(cutout_dimensions, header)
                        logger.debug('Transformed {} into {}'.format(
                            cutout_dimensions, transformed_cutout_dimension))
                        self._pixel_cutout(header, hdu.data,
                                           transformed_cutout_dimension)
                        has_match = True

                except NoContentError:
                    # Skip for now as we're iterating the loop.
                    logger.debug(
                        'No overlap with extension {}'.format(
                            curr_extension_idx))

            logger.debug(
                'Finished extension {}'.format(curr_extension_idx))

        logger.debug('Has match in list? -- {}'.format(has_match))
        return has_match

    def _iterate_hdu_list(self, cutout_dimensions):
        # Start with the first extension
        hdu_list = fits.open(
            name=self.input_stream, memmap=True, mode='readonly',
            do_not_scale_image_data=True)

        # Keep a tally of whether at least one HDU matched.
        has_match = self._check_hdu_list(cutout_dimensions, hdu_list)

        if not has_match:
            raise NoContentError('No overlap found.')

    def _quick_pixel_cutout(self, cutout_dimension):
        hdu_list = fits.open(self.input_stream, memmap=True,
                             mode='readonly', do_not_scale_image_data=True)
        ext_idx = hdu_list.index_of(cutout_dimension.get_extension())
        if ext_idx >= 0:
            hdu = hdu_list.pop(ext_idx)
            self._pixel_cutout(hdu.header, hdu.data, cutout_dimension)

    def _is_single_hdu_cutout(self, cutout_dimensions):
        return cutout_dimensions is not None \
            and len(cutout_dimensions) == 1 \
            and isinstance(cutout_dimensions[0], PixelCutoutHDU)

    def cutout(self, cutout_dimensions):
        if self._is_single_hdu_cutout(cutout_dimensions):
            self._quick_pixel_cutout(cutout_dimensions[0])
        else:
            self._iterate_hdu_list(cutout_dimensions)




__all__ = ['to_num', 'is_integer', 'is_string']


def to_num(s):
    try:
        return int(s)
    except ValueError:
        return float(s)


def is_string(s):
    return s == str(s)


def is_integer(s):
    if isinstance(s, tuple):
        return False

    try:
        int(s)
        return True
    except ValueError:
        return False



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging
import sys
import io
from io import BufferedRandom
import argparse

from cadccutout.utils import is_string
from cadccutout import version
from cadccutout.file_helper import FileHelperFactory
from cadccutout.pixel_range_input_parser import PixelRangeInputParser

logger = logging.getLogger(__name__)

__all__ = ['OpenCADCCutout', 'WriteOnlyStream']


class OpenCADCCutout(object):
    """
    Main cutout class.  This is mainly used as a parent class for concrete
    instances, like from a FITS file, but can be called by itself if need be.

    Parameters
    ----------
    helper_factory : `.file_helper.FileHelperFactory`
        The Helper Factory instance to load a file helper appropriate to the
        file type.  Defaults to file_helper.FileHelperFactory().

    input_range_parser : `.pixel_range_input_parser.PixelRangeInputParser`
        Parser to parse the input string.  This defaults to the provided
        pixel_range_input_parser.PixelRangeInputParser() class.

    Example 1
    --------
    from cadccutout import OpenCADCCutout

    cutout = OpenCADCCutout()
    output_file = tempfile.mkstemp(suffix='.fits')
    input_file = '/path/to/file.fits'

    # Cutouts are in cfitsio format.
    cutout_region_string = '[300:800,810:1000]'  # HDU 0 along two axes.

    # Needs to have 'append' flag set.  The cutout() method will write out the
    # data.
    with open(output_file, 'ab+') as output_writer, open(input_file, 'rb') as
     input_reader:
        test_subject.cutout(input_reader, output_writer, cutout_region_string,
        'FITS')


    Example 2 (CADC)
    --------
    from cadccutout import OpenCADCCutout
    from cadcdata import CadcDataClient

    cutout = OpenCADCCutout()
    anonSubject = net.Subject()
    data_client = CadcDataClient(anonSubject)
    output_file = tempfile.mkstemp(suffix='.fits')
    archive = 'HST'
    file_name = 'n8i311hiq_raw.fits'
    input_stream = data_client.get_file(archive, file_name)

    # Cutouts are in cfitsio format.
    cutout_region_string = '[SCI,10][80:220,100:150]'  # SCI version 10, along
    two axes.

    # Needs to have 'append' flag set.  The cutout() method will write out the
    # data.
    with open(output_file, 'ab+') as output_writer:
        test_subject.cutout(input_stream, output_writer, cutout_region_string,
        'FITS')
    """

    def __init__(self, helper_factory=FileHelperFactory(),
                 input_range_parser=PixelRangeInputParser()):
        self.helper_factory = helper_factory
        self.input_range_parser = input_range_parser

    def cutout(self, cutout_dimensions, input_reader=None,
               output_writer=None, file_type='FITS'):
        """
        Perform a Cutout of the given data at the given position and size.

        Parameters
        ----------
        input_reader: File-like object, Reader stream
            The file location.  The file extension is important as it's used to
            determine how to process it.

        output_writer: File-like object, Writer stream
            The writer to push the cutout array to.

        cutout_dimensions: List of PixelCutoutHDU or WCS Shape objects.
            The requested dimensions expressed as PixelCutoutHDU objects.

        file_type: string
            The file type, in upper case.  Will usually be 'FITS'.
        """

        if not cutout_dimensions or len(cutout_dimensions) == 0:
            raise ValueError('No Cutout regions specified.')

        if input_reader is None:
            # Python 3 uses the buffer property to treat stream data as binary.
            # Python 2 (sometimes) requires the -u command line switch.
            if hasattr(sys.stdin, 'buffer'):
                input_stream = sys.stdin.buffer
            else:
                input_stream = sys.stdin
        else:
            input_stream = input_reader

        if output_writer is None:
            # Python 3 uses the buffer property to treat stream data as binary.
            # Python 2 (sometimes) requires the -u command line switch.
            if hasattr(sys.stdout, 'buffer'):
                output_stream = sys.stdout.buffer
            else:
                output_stream = sys.stdout
        else:
            output_stream = output_writer

        file_helper = self._get_file_helper(
            file_type, input_stream, output_stream)

        try:
            file_helper.cutout(cutout_dimensions)
        except OSError as oe:
            raise ValueError(
                'Output target or input source unusable (Did you specify an '
                'input and output?).\n{}'.format(str(oe)))

    def _parse_input(self, input_cutout_dimensions):
        if self.input_range_parser.is_pixel_cutout(input_cutout_dimensions[0]):
            parsed_cutout_dimensions = self.input_range_parser.parse(
                input_cutout_dimensions[0])
        else:
            # Parse WCS into appropriate objects.
            parsed_cutout_dimensions = input_cutout_dimensions

        return parsed_cutout_dimensions

    def _sanity_check_input(self, cutout_dimensions_str):
        if is_string(cutout_dimensions_str):
            input_cutout_dimensions = [cutout_dimensions_str]
        elif not isinstance(cutout_dimensions_str, list) \
                or not cutout_dimensions_str:
            raise ValueError(
                'Input is expected to be a string or list but was {}'.format(
                    cutout_dimensions_str))
        else:
            input_cutout_dimensions = cutout_dimensions_str

        return input_cutout_dimensions

    def cutout_from_string(self, cutout_dimensions_str, input_reader=None,
                           output_writer=None, file_type='FITS'):
        """
        Perform a Cutout of the given data at the given position and size.

        Parameters
        ----------
        input_reader: File-like object, Reader stream
            The file location.  The file extension is important as it's used to
            determine how to process it.

        output_writer: File-like object, Writer stream
            The writer to push the cutout array to.

        cutout_dimensions_str: string or list of WCS coordinates, or extension
                            and pixel coordinates.  The requested dimensions
                            expressed as PixelCutoutHDU objects.

        file_type: string
            The file type, in upper case.  Will usually be 'FITS'.
        """

        input_cutout_dimensions = self._sanity_check_input(
            cutout_dimensions_str)

        self.cutout(self._parse_input(input_cutout_dimensions), input_reader,
                    output_writer, file_type)

    def _get_file_helper(self, file_type, input_reader, output_writer):
        return self.helper_factory.get_instance(file_type, input_reader,
                                                output_writer)


class WriteOnlyStream(BufferedRandom):
    """
    Stream implementation to seem like a seekable stream.  It is meant to wrap
    the sys.stdout stream so that when Astropy calls the tell() method it
    will have an accurate place to start writing the stream.

    :param raw: file or file-like object.  The Raw underlying stream.
    """

    def __init__(self, raw):
        super(WriteOnlyStream, self).__init__(io.BytesIO())
        self._raw = raw
        self.write_offset = 0
        self.read_offset = 0

    def read(self, size=1):
        raise ValueError('Unreadable stream.  This is write only.')

    def write(self, data):
        written = self._raw.write(data)
        if written:
            self.write_offset += written
        return self.write_offset

    def tell(self):
        return self.write_offset

    def seek(self, offset):
        raise ValueError('Unseekable stream.  This is write only.')


def main_app(argv=None):
    # Execute only if run as a script.
    parser = argparse.ArgumentParser()

    parser.description = ('Cutout library to extract an N-Dimension array.')
    parser.formatter_class = argparse.RawTextHelpFormatter

    parser.add_argument('-d', '--debug', action='store_true',
                        help='debug messages')
    parser.add_argument('-q', '--quiet', action='store_true',
                        help='run quietly')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='verbose messages')
    parser.add_argument('--version', action='version', version=version.version)

    parser.add_argument('--type', '-t', choices=['FITS'], default='FITS',
                        help='Optional file type.  Defaults to FITS.')
    parser.add_argument('--infile', '-i', type=argparse.FileType(mode='rb'),
                        nargs='?', default=None,
                        help='Optional input file.  Defaults to stdin.')
    parser.add_argument('--outfile', '-o', type=argparse.FileType(mode='ab+'),
                        nargs='?', default=None,
                        help='Optional output file.  Defaults to stdout.')

    parser.add_argument(
        'cutout', help='The cutout region string.\n[0][200:400] for a cutout \
        of the 0th extension along the first axis', nargs='+')

    args = parser.parse_args(args=argv)

    if not args:
        parser.print_usage(file=sys.stderr)
        sys.stderr.write("{}: error: too few arguments\n".format(__name__))
        sys.exit(-1)
    if args.verbose:
        level = logging.INFO
    elif args.debug:
        level = logging.DEBUG
    else:
        level = logging.WARN

    logging.basicConfig(level=level)
    logging.getLogger().setLevel(level)

    c = OpenCADCCutout()

    logging.info('Start cutout.')

    # Support multiple strings.  This will write out as many cutouts as
    # it finds.
    c.cutout_from_string(
        args.cutout, input_reader=args.infile,
        output_writer=args.outfile,
        file_type=args.type)


if __name__ == "__main__":
    try:
        main_app()
        sys.exit(0)
    except Exception as e:
        logging.error('{}'.format(str(e)))
        sys.exit(-1)



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)


from cadccutout.cutoutnd import CutoutND
from cadccutout.pixel_range_input_parser import PixelRangeInputParser

__all__ = ['BaseFileHelper']


class BaseFileHelper(object):
    def __init__(self, input_stream, output_writer,
                 input_range_parser=PixelRangeInputParser()):
        """
        Create a new BaseFileHelper used for different file types.  Concrete
        instances are expected to extend this class to provide some common
        state.

        :param input_stream:    The Reader to read the file data from.
        :param output_writer:   The Writer to write the cutout to.
        :param input_range_parser:  Parser instance to parse the range of
        inputs.
        """
        if input_stream is None:
            raise ValueError(
                'An input stream(file-like object or io/stream) is required \
                 to read from.')
        else:
            self.input_stream = input_stream

        if output_writer is None:
            raise ValueError('An output stream(file-like object or io/stream) \
             is required to write to.')
        else:
            self.output_writer = output_writer

        self.input_range_parser = input_range_parser

    def do_cutout(self, data, cutout_dimension, wcs):
        """
        Perform a Cutout of the given data at the given position and size.
        :param data:  The data to cutout from
        :param cutout_dimension:  `PixelCutoutHDU`       Cutout object.
        :param wcs:    The WCS object to use with the cutout to return a copy
         of the WCS object.

        :return: CutoutND instance
        """

        c = CutoutND(data=data, wcs=wcs)
        return c.extract(cutout_dimension)



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging

from math import ceil
from copy import deepcopy

from astropy.wcs import Sip
from astropy.nddata.utils import extract_array
from cadccutout.no_content_error import NoContentError

__all__ = ['CutoutResult', 'CutoutND']

logger = logging.getLogger(__name__)


class CutoutResult(object):
    """
    Just a DTO to move results of a cutout.  It's more readable than a plain
    tuple.
    """

    def __init__(self, data, wcs=None, wcs_crpix=None):
        self.data = data
        self.wcs = wcs
        self.wcs_crpix = wcs_crpix


class CutoutND(object):
    def __init__(self, data, wcs=None):
        """
        Parameters
        ----------
        data : `~numpy.ndarray`
            The N-dimensional data array from which to extract the cutout array.
        cutout_region : `PixelCutoutHDU`
            The Pixel HDU Cutout description.  See
            cadccutout.pixel_cutout_hdu.py.
        wcs : `~astropy.wcs.WCS` or `None`
            A WCS object associated with the cutout array.  If it's specified,
            reset the WCS values for the cutout.

        Returns
        -------
        CutoutResult instance
        """
        if data is None:
            raise NoContentError('Nothing to cutout from.')

        self.data = data
        self.wcs = wcs

    def _get_position_shape(self, data_shape, cutout_region):
        requested_shape = cutout_region.get_shape()
        requested_position = cutout_region.get_position()

        # reverse position because extract_array uses reverse ordering
        # (i.e. x,y -> y,x).
        r_position = tuple(reversed(requested_position))
        r_shape = tuple(reversed(requested_shape))

        len_data = len(data_shape)
        len_pos = len(r_position)
        len_shape = len(r_shape)

        logger.debug('Data shape: {}'.format(data_shape))
        logger.debug('Requested shape: {}'.format(r_shape))
        logger.debug('Requested position: {}'.format(r_position))

        if len_shape > len_data:
            raise ValueError('Invalid shape requested (tried to extract {} \
from {}).'.format(r_shape, data_shape))

        shape = (data_shape[:(len_data - len_shape)]) + r_shape
        prepend_list = data_shape[:(len_data - len_pos)]
        prepend_position_list = []

        for idx, val in enumerate(prepend_list):
            prepend_position_list.append(int(ceil((val / 2) - 0.5)))

        position = tuple(prepend_position_list) + r_position

        return (position, shape)

    def extract(self, cutout_region):
        data = self.data
        data_shape = data.shape
        position, shape = self._get_position_shape(data_shape, cutout_region)
        logger.debug('Position {} and Shape {}'.format(position, shape))

        # No pixels specified, so return the entire HDU
        if (not position and not shape) or shape == data_shape:
            logger.debug('Returning entire HDU data for {}'.format(
                cutout_region.get_extension()))
            cutout_data = data
        else:
            logger.debug('Cutting out {} at {} for extension {} from  \
            {}.'.format(
                shape, position, cutout_region.get_extension(), data.shape))
            cutout_data, position = extract_array(
                data, shape, position, mode='trim', return_position=True)

        if self.wcs:
            cutout_shape = cutout_data.shape
            output_wcs = deepcopy(self.wcs)
            wcs_crpix = output_wcs.wcs.crpix
            l_wcs_crpix = len(wcs_crpix)
            ranges = cutout_region.get_ranges()

            logger.debug('Adjusting WCS.')

            for idx, _ in enumerate(ranges):
                if idx < l_wcs_crpix:
                    wcs_crpix[idx] -= (ranges[idx][0] - 1.0)

            output_wcs._naxis = list(cutout_shape)

            if self.wcs.sip:
                curr_sip = self.wcs.sip
                output_wcs.sip = Sip(curr_sip.a, curr_sip.b,
                                     curr_sip.ap, curr_sip.bp,
                                     wcs_crpix[0:2])

            logger.debug('WCS adjusted.')
        else:
            logger.debug('No WCS present.')
            output_wcs = None
            wcs_crpix = None

        return CutoutResult(data=cutout_data, wcs=output_wcs,
                            wcs_crpix=wcs_crpix)



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#

import logging
import numpy as np

from math import ceil
from cadccutout.utils import is_integer

__all__ = ['PixelCutoutHDU']


logger = logging.getLogger(__name__)


def fix_tuple(t):
    if np.isscalar(t):
        return (t, t)
    elif len(t) < 2:
        raise ValueError('Unusable dimension range {}'.format(t))
    else:
        return t


class PixelCutoutHDU(object):
    def __init__(self, dimension_ranges=[], extension='0'):
        """
        A Pixel cutout.
        :param dimension_ranges: list    Dimension ranges expressed as tuples
        (i.e. (lower,upper)).
        :param extension: tuple, int, string
            The Extension specification to use.  If tuple, use (str, int) to
             get the nth count of the EXTNAME=str
            extension.  If string, use the first extension with EXTNAME=string,
             or use int to get the extension[int].
            This is zero (0) based.
        """
        self.dimension_ranges = list(map(fix_tuple, dimension_ranges))
        self._extension = str(extension)  # For consistency.
        logger.debug(
            'Requested dimension ranges ({}) for extension {}.'.format(
                self.dimension_ranges, self._extension))

    def get_ranges(self):
        """
        Obtain the range tuples.
        """
        acc = []
        for range_tuple in self.dimension_ranges:
            acc.append(
                (int(np.round(range_tuple[0])), int(np.round(range_tuple[1]))))

        return acc

    def get_shape(self):
        """
        Convert the given dimensions to a shape.
        """
        acc = []
        for range_tuple in self.dimension_ranges:
            acc.append(int(np.round((range_tuple[1] - range_tuple[0]) + 1)))

        return tuple(acc)

    def get_position(self):
        """
        Convert the given dimensions to a position to extract.
        """
        acc = []
        for range_tuple in self.dimension_ranges:
            acc.append(int(ceil(
                range_tuple[0] - 0.5) + int(ceil(((range_tuple[1]
                                                   - range_tuple[0]) / 2)
                                                 - 0.5))) - 1)

        return tuple(acc)

    def get_extension(self):
        ext = self._extension
        if is_integer(ext):
            return int(ext)
        elif ext.count(',') == 1:
            es = ext.split(',')
            ext_int = int(es[1])

            # EXTNAME and EXTVER are 1-based.
            if ext_int == 0:
                ext_int = 1

            return (es[0], ext_int)
        else:
            return (ext, 1)

    def __str__(self):
        return 'PixelCutoutHDU(dimensions={}, extension={})'.format(
            self.dimension_ranges, self._extension)



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging
import math
import sys

from astropy import units as u
from astropy.coordinates import SkyCoord, Longitude, Latitude, ICRS
from astropy.wcs import WCS, NoConvergence
from regions.shapes.circle import CircleSkyRegion
from regions.shapes.polygon import PolygonSkyRegion

from .no_content_error import NoContentError
from .pixel_cutout_hdu import PixelCutoutHDU
from .shape import Circle, Polygon, Energy, Time, Polarization, \
    PolarizationState

logger = logging.getLogger(__name__)

__all__ = ['AxisType', 'Transform']

_DEFAULT_ENERGY_CTYPE = 'WAVE-???'
_DEFAULT_ENERGY_CUNIT = 'm'


class AxisType(object):
    """
    Extracts the axis number for each coordinate type.
    """

    def _get_wcs(self, header):
        naxis_value = header.get('NAXIS')

        if naxis_value is not None and int(naxis_value) > 0:
            for nax in range(1, naxis_value + 1):
                ctype = header.get('CTYPE{0} '.format(nax))
                if ctype is not None and ctype.endswith('-SIP'):
                    naxis = 2
                    break
                else:
                    naxis = None
        else:
            naxis = naxis_value

        return WCS(header=header, naxis=naxis, fix=False)

    COORDINATE_TYPE = 'coordinate_type'
    SPATIAL_KEYWORDS = ['celestial']
    SPECTRAL_KEYWORDS = ['spectral']
    TEMPORAL_KEYWORDS = []
    POLARIZATION_KEYWORDS = ['stokes']

    def __init__(self, header):
        self.spatial_1 = None
        self.spatial_2 = None
        self.spectral = None
        self.temporal = None
        self.polarization = None

        # wcs from header
        wcs = self._get_wcs(header)

        # get list of dict of axis types
        axis_types = wcs.get_axis_types()

        # for each coordinate dict extract type and axis number
        for i in range(len(axis_types)):
            axis_type = axis_types[i]
            coordinate_type = axis_type.get(self.COORDINATE_TYPE)
            if coordinate_type in self.SPATIAL_KEYWORDS:
                if not self.spatial_1:
                    self.spatial_1 = i + 1
                else:
                    self.spatial_2 = i + 1
                    logger.info('Spatial naxis{} naxis{}'.format(
                        self.spatial_1, self.spatial_2))
            elif coordinate_type in self.SPECTRAL_KEYWORDS:
                self.spectral = i + 1
                logger.info('Spectral naxis{}'.format(
                    self.get_spectral_axis()))
            elif coordinate_type in self.TEMPORAL_KEYWORDS:
                self.temporal = i + 1
                logger.info('Temporal naxis{}'.format(
                    self.get_temporal_axis()))
            elif coordinate_type in self.POLARIZATION_KEYWORDS:
                self.polarization = i + 1
                logger.info('Polarization naxis{}'.format(
                    self.get_polarization_axis()))
            elif coordinate_type:
                raise ValueError('Unknown axis keyword {}'.format(
                    coordinate_type))

    def get_spatial_axes(self):
        """
        Get a list of the spatial axis numbers.
        :return: List[int], the two spatial axis numbers,
                            or None if no spatial axes found.
        """
        return [self.spatial_1, self.spatial_2]

    def get_spectral_axis(self):
        """
        Get the spectral axis number.
        :return: int, the spectral axis number,
                      or None if no spectral axis found.
        """
        return self.spectral

    def get_temporal_axis(self):
        """
        Get the temporal axis number.
        :return: int, the temporal axis number,
                      or None if no temporal axis found.
        """
        return self.temporal

    def get_polarization_axis(self):
        """
        Get the polarization axis number.
        :return: int, the polarization axis number,
                      or None if no polarization axis found.
        """
        return self.polarization


class Transform(object):

    def _append_world_to_circle_pixels(self, header, naxis1, naxis2, pixels,
                                       cutouts):
        logger.info('CIRCLE pixels [{}:{}, {}:{}]'.format(*pixels))

        l_cutouts = len(cutouts)

        for i in pixels:
            if not i:
                logger.debug('Skipping {}'.format(header))
                raise NoContentError(
                    'Unable to create bounding box.')

        logger.debug('Removing {} from {}'.format(naxis1, cutouts))

        # remove default cutouts and add query cutout
        if naxis1 < l_cutouts:
            cutouts.pop(naxis1)
            cutouts.insert(naxis1, (pixels[0], pixels[1]))

        if naxis2 < l_cutouts:
            cutouts.pop(naxis2)
            cutouts.insert(naxis2, (pixels[2], pixels[3]))

    def world_to_pixels(self, world_coords, header):
        """
        Convert a list of world coordinates to pixel coordinates
        for the given FITS header.

        :param world_coords: List List of world coordinate strings
        :param header: Header   The FITS header
        :return: PixelCutoutHDU containing the pixel coordinates
        """
        logger.info('Shapes: {}'.format(world_coords))

        # coordinate axis numbers
        axis_types = AxisType(header)

        # get the Shape object for the world coordinates
        shapes = self.world_to_shape(world_coords)

        # list of NAXIS length, with 1 to NAXIS[1|2|3|4] pixel coordinates,
        # add a dummy object to the start of the list so the axes
        # align with the list index, remove it at the end.
        cutouts = [(0, 0)]
        axes = header.get('NAXIS')
        for i in range(1, axes + 1):
            length = header.get('NAXIS{}'.format(i))
            cutouts.append((1, length))

        # accumulated NoContentErrors
        no_content_errors = []

        # get the pixel coordinates for each type of shape
        # do try except for each call to catch NoContentError for no overlap
        # raise no content error if an axis has no overlap
        for shape in shapes:
            if isinstance(shape, Circle):
                try:
                    # length of the two axes
                    naxis1 = axis_types.get_spatial_axes()[0]
                    naxis2 = axis_types.get_spatial_axes()[1]
                    # get the cutout pixels
                    pixels = self.get_circle_cutout_pixels(
                        shape, header, naxis1, naxis2)
                    self._append_world_to_circle_pixels(header, naxis1, naxis2,
                                                        pixels, cutouts)
                except NoContentError as e:
                    no_content_errors.append(repr(e))
            elif isinstance(shape, Polygon):
                try:
                    # length of the two axes
                    naxis1 = axis_types.get_spatial_axes()[0]
                    naxis2 = axis_types.get_spatial_axes()[1]

                    # get the cutout pixels
                    pixels = self.get_polygon_cutout_pixels(
                        shape, header, naxis1, naxis2)
                    logger.info(
                        'POLYGON pixels [{}:{}, {}:{}]'.format(*pixels))

                    # remove default cutouts and add query cutout
                    cutouts.pop(naxis1)
                    cutouts.insert(naxis1, (pixels[0], pixels[1]))
                    cutouts.pop(naxis2)
                    cutouts.insert(naxis2, (pixels[2], pixels[3]))
                except NoContentError as e:
                    no_content_errors.append(repr(e))
            elif isinstance(shape, Energy):
                try:
                    # length of the spectral axis
                    naxis = axis_types.get_spectral_axis()

                    # get the cutout pixels
                    pixels = self.get_energy_cutout_pixels(shape, header, naxis)
                    logger.info('BAND pixels [{}:{}]'.format(*pixels))

                    # remove default cutouts and add query cutout
                    cutouts.pop(naxis)
                    cutouts.insert(naxis, (pixels[0], pixels[1]))
                except NoContentError as e:
                    no_content_errors.append(repr(e))
            elif isinstance(shape, Time):
                try:
                    # length of the temporal axis
                    naxis = axis_types.get_temporal_axis()

                    # get the cutout pixels
                    pixels = self.get_time_cutout_pixels(shape, header, naxis)
                    logger.info('TIME pixels [{}:{}]'.format(*pixels))

                    # remove default cutouts and add query cutout
                    cutouts.pop(naxis)
                    cutouts.insert(naxis, (pixels[0], pixels[1]))
                except NoContentError as e:
                    no_content_errors.append(repr(e))
            elif isinstance(shape, Polarization):
                try:
                    # length of the polarization axis
                    naxis = axis_types.get_polarization_axis()

                    # get the cutout pixels
                    pixels = self.get_polarization_cutout_pixels(
                        shape, header, naxis)
                    logger.info('POL pixels [{}:{}]'.format(*pixels))

                    # remove default cutouts and add query cutout
                    cutouts.pop(naxis)
                    cutouts.insert(naxis, (pixels[0], pixels[1]))
                except NoContentError as e:
                    no_content_errors.append(repr(e))

        # check for no content errors
        if no_content_errors:
            raise NoContentError('\n'.join(no_content_errors))

        # remove the dummy first list item
        cutouts.pop(0)

        return PixelCutoutHDU(cutouts)

    def world_to_shape(self, world_coords):
        """
        Parse the world coordinates into a list of Shapes.

        :param world_coords: List List of world coordinate strings.
        :return: List A list of Shapes.
        """

        # list of Shape
        shapes = []

        # track names of shapes to check for duplicates
        shape_types = []
        for world in world_coords:

            # get the Shape for this world coordinate
            shape = self.parse_world(world)

            # check for duplicate shapes, one shape per axis allowed
            if shape.AXIS_TYPE in shape_types:
                raise ValueError('One cutout per axis supported, '
                                 'found duplicate world axis for {}'
                                 .format(shape.NAME))

            shapes.append(shape)
            shape_types.append(shape.AXIS_TYPE)

        return shapes

    def parse_world(self, world):
        """

            :param world: str   World coordinate string.
            :returns:
        """

        if not world.strip():
            raise ValueError('Empty or None world coordinate')

        parameters = world.upper().split()
        name = parameters[0].strip()
        values = list(filter(None, parameters[1:]))
        values = ' '.join(values)

        if name == Circle.NAME:
            return Circle.parse_circle(values)
        elif name == Polygon.NAME:
            return Polygon.parse_polygon(values)
        elif name == Energy.NAME:
            return Energy.parse_energy(values)
        elif name == Time.NAME:
            return Time.parse_time(values)
        elif name == Polarization.NAME:
            return Polarization.parse_polarization(values)
        else:
            raise ValueError('Unknown world coordinate {}'.format(world))

    def get_circle_cutout_pixels(self, circle, header, naxis1, naxis2):
        """
        Get the pixels coordinates for a Circle query
        using the given FITS header.

        :param circle: Circle Circle instance
        :param header: Header   FITS extension header
        :param naxis1: int  First spatial axis
        :param naxis2: int  Second spatial axis
        :return: List[int] The x and y pairs of the pixel coordinates
        """

        # sky coordinates from the Circle
        ra = Longitude(circle.ra, unit=u.deg)
        dec = Latitude(circle.dec, unit=u.deg)
        radius = u.Quantity(circle.radius, unit=u.deg)
        sky_coords = SkyCoord(ra, dec, frame=ICRS)

        # WCS from the header, extract only the spatial axes wcs, the sky to pix
        # transform will want to convert each axis in the wcs, and we only have
        # spatial data.
        wcs = WCS(header, naxis=[naxis1, naxis2], fix=False)

        # Circle region with radius
        sky_region = CircleSkyRegion(sky_coords, radius=radius)

        try:
            # convert to pixel coordinates
            pixels = sky_region.to_pixel(wcs)

            x_min = pixels.bounding_box.ixmin
            x_max = pixels.bounding_box.ixmax
            y_min = pixels.bounding_box.iymin
            y_max = pixels.bounding_box.iymax
            logger.debug('Circle bounding box [{}, {}, {}, {}]'.format(
                x_min, x_max, y_min, y_max))
        except ValueError as e:
            # bounding_box raises ValueError if the cutout
            # doesn't intersect the image
            raise NoContentError(repr(e))
        except NoConvergence as nce:
            raise NoContentError(repr(nce))

        # do clip check
        x_axis = header.get('NAXIS{}'.format(naxis1))
        y_axis = header.get('NAXIS{}'.format(naxis2))
        return self.do_position_clip_check(
            x_axis, y_axis, x_min, x_max, y_min, y_max)

    def get_polygon_cutout_pixels(self, polygon, header, naxis1, naxis2):
        """
        Get the pixels coordinates for a Circle query
        using the given FITS header.

        :param polygon: Polygon  Polygon instance
        :param header: Header   FITS extension header
        :param naxis1: int  First spatial axis
        :param naxis2: int  Second spatial axis
        :return: List[int] The x and y pairs of the cutout pixel coordinates
        """

        # query sky position
        even_vertices = []
        odd_vertices = []
        for vertice in polygon.vertices:
            even_vertices.append(vertice[0])
            odd_vertices.append(vertice[1])
        sky_coords = SkyCoord(even_vertices, odd_vertices,
                              unit=u.deg, frame=ICRS)

        # WCS from the header, extract only the spatial axes,
        # because the sky to pix transform will want to convert
        # every axis in the wcs, and we only have spatial data to convert.
        wcs = WCS(header, naxis=[naxis1, naxis2], fix=False)

        # Polygon region
        sky_region = PolygonSkyRegion(sky_coords)

        # convert to pixel coordinates
        pixels = sky_region.to_pixel(wcs)
        try:
            x_min = pixels.bounding_box.ixmin
            x_max = pixels.bounding_box.ixmax
            y_min = pixels.bounding_box.iymin
            y_max = pixels.bounding_box.iymax
            logger.info('Polygon bounding box [{}, {}, {}, {}]'.format(
                x_min, x_max, y_min, y_max))
        except ValueError as e:
            # bounding_box raises ValueError if the cutout
            # doesn't intersect the image
            raise NoContentError(repr(e))

        # do clip check
        x_axis = header.get('NAXIS{}'.format(naxis1))
        y_axis = header.get('NAXIS{}'.format(naxis2))
        return self.do_position_clip_check(x_axis, y_axis,
                                           x_min, x_max, y_min, y_max)

    def get_energy_cutout_pixels(self, energy, header, naxis):
        """
        Get the pixels coordinates for a spectral query
        using the given FITS header.

        :param energy: Energy   Energy instance
        :param header: Header   FITS extension header
        :param naxis: int   Spectral axis number
        :return: List[int] The two cutout pixel coordinates
        """

        # WCS from the header
        wcs = WCS(header, fix=False)

        # if the spectral wcs isn't a wavelength, transform the
        # spectral axis to wavelength.
        ctype = header.get('CTYPE{}'.format(naxis))
        if ctype and not ctype.startswith('WAVE'):
            try:
                wcs.wcs.sptr(_DEFAULT_ENERGY_CTYPE)
            except ValueError as e:
                error = 'wcslib error transforming from {} to {}: {}'\
                    .format(ctype, _DEFAULT_ENERGY_CTYPE, repr(e))
                raise ValueError(error)

        # slice out the spectral wcs
        energy_wcs = wcs.sub([naxis])

        # convert from world to pixel coordinates
        pixels = energy_wcs.wcs_world2pix([energy.lower, energy.upper], 1)

        # clip check
        length = header.get('NAXIS{}'.format(naxis))
        lower = min(pixels[0][0], pixels[0][1])
        upper = max(pixels[0][0], pixels[0][1])
        logger.info('Energy interval: {}:{}'.format(lower, upper))

        return self.do_energy_clip_check(length, lower, upper)

    def get_time_cutout_pixels(self, time, header, naxis):
        """
        Get the pixels coordinates for a temporal query
        using the given FITS header.

        :param time: Time  Time instance
        :param header: Header   FITS extension header
        :param naxis: int   Temporal axis number
        :return: List[int] The two cutout pixel coordinates
        """

        raise ValueError('Temporal cutouts not supported')

    def get_polarization_cutout_pixels(self, polarization, header, naxis):
        """
        Get the pixels coordinates for a polarization query
        using the given FITS header.

        :param polarization: Polarization Polarization instance
        :param header: Header   FITS extension header
        :param naxis: int   Polarization axis number
        :return: List[int] The two cutout pixel coordinates
        """

        crval = header.get('CRVAL{}'.format(naxis))
        crpix = header.get('CRPIX{}'.format(naxis))
        cdelt = header.get('CDELT{}'.format(naxis))

        if not crval and not crpix and not cdelt:
            raise ValueError('Unable to do a polarization cutout, '
                             'incomplete WCS, missing one of {} {} {}'
                             .format('CRVAL{}'.format(naxis), 'CRPIX{}'
                                     .format(naxis), 'CDELT{}'.format(naxis)))

        # polarization states in the file
        wcs_states = self.get_wcs_polarization_states(header, naxis)
        if not wcs_states:
            raise NoContentError('Polarization states not found '
                                 'in the FITS header')

        # list of states in both the cutout query and header
        cutout_states = []
        for state in polarization.states:
            if state in wcs_states:
                cutout_states.append(state)

        # no overlap, raise NoContentError
        if not cutout_states:
            raise NoContentError('No polarization states in the query {} '
                                 'match the states in the FITS header {}'
                                 .format(str(polarization.states),
                                         str(wcs_states)))

        # calculate pixels for the states to cutout
        pix1 = sys.maxsize
        pix2 = -sys.maxsize - 1
        for state in cutout_states:
            value = PolarizationState(state).value
            pix = crpix + (value - crval) / cdelt
            pix1 = min(pix1, pix)
            pix2 = max(pix2, pix)
        logger.info('Polarization states: {}:{}'.format(pix1, pix2))

        # do clip check
        pixels = self.do_polarization_clip_check(naxis, pix1, pix2)
        return pixels

    def do_position_clip_check(self, w, h, x1, x2, y1, y2):
        """
        Trim the cutout coordinates to fit within the image bounds.

        :param w: int   Width of the image
        :param h: int   Height of the image
        :param x1: int  Lower x cutout coordinate
        :param x2: int  Upper x cutout coordinate
        :param y1: int  Lower y cutout coordinate
        :param y2: int  Upper y cutout coordinate
        :return: List[int] The coordinates pixels within the images bounds
        """

        # bounds check
        if x1 < 1:
            x1 = 1
        if w and x2 > w:
            x2 = w
        if y1 < 1:
            y1 = 1
        if h and y2 > h:
            y2 = h

        # cutout pixels
        logger.info('Position clip check: [{}, {}, {}, {}]'.format(
            x1, x2, y1, y2))
        return [x1, x2, y1, y2]

    def do_energy_clip_check(self, naxis, lower, upper):
        """
        Trim the cutout coordinates to fit within the spectral axis.

        :param naxis: int   Length of the spectral axis
        :param lower: float Lower cutout bounds
        :param upper: float Upper cutout bounds
        :return: List[int]  The coordinate pixels within the spectral axis.
        """

        # round floats to individual pixels
        z1 = math.floor(lower)
        z2 = math.ceil(upper)

        # bounds check
        if z1 < 1:
            z1 = 1
        if z2 > naxis:
            z2 = naxis

        # validity check, no pixels included
        if z1 >= naxis or z2 <= 1:
            error = 'pixels coordinates {}:{} do not intersect {} to {}'\
                .format(z1, z2, 1, naxis)
            raise NoContentError(error)

        # an actual cutout
        logger.info('Energy clip check: [{}:{}]'.format(z1, z2))
        return [z1, z2]

    def do_time_clip_check(self, naxis, lower, upper):
        """
        Trim the cutout coordinates to fit within the temporal axis.

        :param naxis: int   Length of the temporal axis
        :param lower: float Lower cutout bounds
        :param upper: float Upper cutout bounds
        :return: List[int]  The coordinate pixels within the temporal axis.
        """

        return []

    def do_polarization_clip_check(self, naxis, lower, upper):
        """
        Trim the cutout coordinates to fit within the polarization axis.

        :param naxis: int   Length of the polarization axis
        :param lower: float Lower cutout bounds
        :param upper: float Upper cutout bounds
        :return: List[int]  The coordinate pixels within the polarization axis.
        """

        # round floats to individual pixels
        p1 = math.floor(lower)
        p2 = math.ceil(upper)

        # bounds check
        if p1 < 1:
            p1 = 1
        if p2 > naxis:
            p2 = naxis

        # validity check, no pixels included
        if p1 > naxis or p2 < 1:
            error = 'pixels coordinates {}:{} do not intersect {} to {}'\
                .format(p1, p2, 1, naxis)
            raise NoContentError(error)

        # an actual cutout
        return [p1, p2]

    def get_wcs_polarization_states(self, header, naxis):
        """
        Get the polarization states from the FITS header.

        :param header: Header   The FITS header
        :param naxis: int   The polarization axis number
        :return: List[PolarizationState]    List of PolarizationState
                                            in the header
        """

        crval = header.get('CRVAL{}'.format(naxis))
        crpix = header.get('CRPIX{}'.format(naxis))
        cdelt = header.get('CDELT{}'.format(naxis))

        polarization_states = []
        if crval and crpix and cdelt:
            for i in range(1, naxis + 1):
                pix = crpix + (i - crval) / cdelt
                polarization_states.append(PolarizationState(pix))

        return polarization_states
